# -*- coding: utf-8 -*-
"""
Created on Wed Sep 16 15:15:30 2020

@author: yhh18
@desc: 公式测试environment
"""
from PC_env import radio_environment
import math


class TestEnvFunc:

    def __init__(self):
        self.seed = 2  # Seed can take any value between 0 and 49
        self.env = radio_environment(self.seed)

    def test_path_loss_sub6(self):
        # The user and base station are in the same location.
        path_loss_1 = self.env._path_loss_sub6(0, 0)
        assert path_loss_1 == 0
        # The distance between user and base station is 5m.
        path_loss_2 = self.env._path_loss_sub6(3, 4)
        path_loss_2 = float(format(path_loss_2, '.6f'))
        assert path_loss_2 == 134.181195

    def test_compute_bf_vector(self):
        bf_vector_1 = self.env._compute_bf_vector(math.pi/3)
        bf_vector_2 = [(0.5+0.0j), (-1.9142843494634747e-16+0.5j),
                       (-0.5-3.8285686989269494e-16j), (7.963299097640738e-16-0.5j)]
        assert not (bf_vector_2 - bf_vector_1).any()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
@author: Jiawei Wu
@create time: 2020-09-25 11:20
@edit time: 2020-11-04 11:21
@FilePath: /vvlab/vvlab/envs/power_allocation/pa_env.py
@desc: An enviornment for power allocation in d2d and BS het-nets.

Path_loss is 114.8 + 36.7*np.log10(d), follow 3GPP TR 36.873, d is
the transmitting distance, fc is 3.5GHz.
Bandwith is 20MHz, AWGN power is -114dBm, respectively.
Assume BS power is lower than 46 dBm(about 40 W).
Assume minimum transmit power: 5dBm/ maximum: 38dBm, for d2d devices.
FP algorithm, WMMSE algorithm, maximum power, random power allocation
schemes as comparisons.
downlink
"""
from collections import namedtuple
from pathlib import Path

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.io
import scipy.special

Node = namedtuple('Node', 'x y type')


class PAEnv:
    @property
    def n_states(self):
        """return dim of state"""
        part_len = {
            'power': self.m_state + 1,
            'rate': self.m_state + 1,
            'fading': self.m_state
        }
        return sum(part_len[metric] for metric in self.metrics)

    @property
    def n_actions(self):
        """return num of actions"""
        return self.n_levels

    def init_observation_space(self, kwargs):
        valid_parts = ['power', 'rate', 'fading']
        # default using power
        sorter = kwargs['sorter'] if 'sorter' in kwargs else 'power'
        # default using all
        metrics = kwargs['metrics'] if 'metrics' in kwargs else valid_parts

        # check data
        if sorter not in valid_parts:
            msg = f'sorter should in power, rate and fading, but is {sorter}'
            raise ValueError(msg)
        if any(metric not in valid_parts for metric in metrics):
            msg = f'metrics should in power, rate and fading, but is {metrics}'
            raise ValueError(msg)

        # set to instance attr
        self.sorter, self.metrics = sorter, metrics

    def init_power_levels(self):
        min_power, max_power = self.min_power, self.max_power
        zero_power = 0
        def cal_p(p_dbm): return 1e-3 * np.power(10, p_dbm / 10)
        dbm_powers = np.linspace(min_power, max_power, num=self.n_levels-1)
        powers = [cal_p(p_dbm) for p_dbm in dbm_powers]
        powers = [zero_power] + powers
        self.power_levels = np.array(powers)

    def init_pos(self):
        """ Initialize position of devices(DT, DR, BS and CUE).

        Establish a Cartesian coordinate system using the BS as an origin.

        Celluar User Equipment(CUE) are all located within the radius R_bs
        of the base station(typically, 1km) and outside the protected radius
        r_bs(simulation paramter, typically 0.01km) of the BS.

        The D2D transmission devices(DT) are located within the radius
        (R_bs - R_dev) of the BS, to ensure DRs all inside the radius R_bs.
        Each DT has a cluster with sveral DRs, which is allocated within the
        radius R_dev of DT. Futher more, DRs always appear outside the raduis
        r_dev(typically 0.001km) of its DT.

        All positions are sotred with the infomation of the corresponding
        device in attributes of the environment instance, self.users and
        self.devices.
        """
        r_bs, R_bs, r_dev, R_dev = self.r_bs, self.R_bs, self.r_dev, self.R_dev

        def random_point(min_r, radius, ox=0, oy=0):
            # https://www.cnblogs.com/yunlambert/p/10161339.html
            # renference the formulaic deduction, his code has bug at uniform
            theta = np.random.random() * 2 * np.pi
            r = np.random.uniform(min_r, radius**2)
            x, y = np.cos(theta) * np.sqrt(r), np.sin(theta) * np.sqrt(r)
            return ox + x, oy + y
        # init CUE positions
        self.station = Node(0, 0, 'station')
        self.users = {}
        for i in range(self.m_usr):
            x, y = random_point(r_bs, R_bs)
            user = Node(x, y, 'user')
            self.users[i] = user

        # init D2D positions
        self.devices = {}
        for t in range(self.n_t):
            tx, ty = random_point(r_bs, R_bs - R_dev)
            t_device = Node(tx, ty, 't_device')
            r_devices = {
                r: Node(*random_point(r_dev, R_dev, tx, ty), 'r_device')
                for r in range(self.m_r)
            }
            self.devices[t] = {'t_device': t_device, 'r_devices': r_devices}

    def init_jakes(self, fd=10, Ts=20e-3, Ns=50):
        """Initialize samples of Jakes model.

        Jakes model is a simulation of the rayleigh channel, which represents
        the small-scale fading.

        Each Rx corresponding to a (downlink) channel, each channel is a
        source of interference to other channels. Consdering the channel
        itself, we get a matrix representing the small-scale fading. Note that
        the interference is decided on the position of Tx and Rx, so that the
        m interferences make by m channels from the same Tx have the same
        fading ratio.

        Args:
            fd: Doppler frequency, default 10(Hz)
            Ts: sampling period, default 20 * 1e-3(s)
            Ns: number of samples, default 50
        """
        n_t, m_r, n_bs, m_usr = self.n_t, self.m_r, 1, self.m_usr
        n_recvs = n_t * m_r + n_bs * m_usr
        randn = np.random.randn

        def calc_h_set(pho):
            # calculate next sample of Jakes model.
            h_d2d = np.kron(np.sqrt((1.-pho**2)*0.5*(
                randn(n_recvs, n_t)**2 + randn(n_recvs, n_t)**2)),
                np.ones((1, m_r), dtype=np.int32))
            h_bs = np.kron(np.sqrt((1.-pho**2)*0.5*(
                randn(n_recvs, n_bs)**2 + randn(n_recvs, n_bs)**2)),
                np.ones((1, m_usr), dtype=np.int32))
            h_set = np.concatenate((h_d2d, h_bs), axis=1)
            return h_set
        # recurrence generate all Ns samples of Jakes.
        H_set = np.zeros([n_recvs, n_recvs, int(Ns)], dtype=np.float32)
        pho = np.float32(scipy.special.k0(2*np.pi*fd*Ts))
        H_set[:, :, 0] = calc_h_set(0)
        for i in range(1, int(Ns)):
            H_set[:, :, i] = H_set[:, :, i-1]*pho + calc_h_set(pho)

        self.H_set = H_set

    def init_path_loss(self, slope=0):
        """Initialize paht loss( large-scale fading).

        The large-scale fading is related to distance. An experimental
        formula can be used to modelling it by 3GPP TR 36.873, explained as:
        L = 36.7log10(d) + 22.7 + 26log10(fc) - 0.3(hUT - 1.5).
        When fc=3.5GHz and hUT=1.5m, the formula can be simplified to:
        L = 114.8 + 36.7*log10(d) + 10*log10(z),
        where z is a lognormal random variable.

        As with the small-scale fading, each the n Rxs have one siginal and
        (n-1) interferences.  Using a n*n matrix to record the path loss, we
        notice that the interference from one Tx has same large-scale fading,
        Consistent with small-scale fading.
        """
        n_t, m_r, n_bs, m_usr = self.n_t, self.m_r, self.n_bs, self.m_usr
        n_r_devices, n_recvs = n_t * m_r, n_t * m_r + n_bs * m_usr

        # calculate distance matrix from initialized positions.
        distance_matrix = np.zeros((n_recvs, n_recvs))

        def get_distances(node):
            """Calculate distances from other devices to given device."""
            dis = np.zeros(n_recvs)
            # d2d
            for t_index, cluster in self.devices.items():
                t_device = cluster['t_device']
                delta_x, delta_y = t_device.x - node.x, t_device.y - node.y
                distance = np.sqrt(delta_x**2 + delta_y**2)
                dis[t_index*m_r: t_index*m_r+m_r] = distance
            # bs
            delta_x, delta_y = self.station.x - node.x, self.station.y - node.y
            distance = np.sqrt(delta_x**2 + delta_y**2)
            dis[n_r_devices:] = distance    # 已经有n_r_devices个信道了
            return dis

        # 接收器和干扰项都先考虑d2d再考虑基站
        for t_index, cluster in self.devices.items():
            r_devices = cluster['r_devices']
            for r_index, r_device in r_devices.items():
                distance_matrix[t_index * self.m_r +
                                r_index] = get_distances(r_device)
        for u_index, user in self.users.items():
            distance_matrix[n_r_devices + u_index] = get_distances(user)

        self.distance_matrix = distance_matrix

        # assign the minimum distance
        min_dis = np.concatenate(
            (np.repeat(self.r_dev, n_r_devices), np.repeat(self.r_bs, m_usr))
        ) * np.ones((n_recvs, n_recvs))
        std = 8. + slope * (distance_matrix - min_dis)
        # random initialize lognormal variable
        lognormal = np.random.lognormal(size=(n_recvs, n_recvs), sigma=std)

        # micro
        path_loss = lognormal * \
            pow(10., -(114.8 + 36.7*np.log10(distance_matrix))/10.)
        self.path_loss = path_loss

    def __init__(self, n_levels,
                 n_t_devices=9, m_r_devices=4, n_bs=1, m_usrs=4, **kwargs):
        """Initialize PA environment"""
        # set sttributes
        self.n_t, self.m_r = n_t_devices, m_r_devices
        self.n_bs, self.m_usr = n_bs, m_usrs
        self.n_recvs = self.n_t * self.m_r + self.n_bs * self.m_usr
        self.r_dev, self.r_bs, self.R_dev, self.R_bs = 0.001, 0.01, 0.1, 1
        self.Ns, self.n_levels = 50, n_levels
        self.min_power, self.max_power, self.thres_power = 5, 38, -114  # dBm
        self.bs_power = 10  # W
        self.m_state = 16
        self.__dict__.update(kwargs)
        # set random seed
        if 'seed' in kwargs:
            seed = kwargs['seed'] if kwargs['seed'] > 1 else 799345
            np.random.seed(seed)
            print(f'PAEnv set random seed {seed}')

        # init attributes of pa env
        self.init_observation_space(kwargs)
        self.init_power_levels()
        self.init_pos()  # init recv pos
        self.init_jakes(Ns=self.Ns)  # init rayleigh loss using jakes model
        self.init_path_loss(slope=0)  # init path loss
        self.cur_step = 0

    def reset(self):
        self.cur_step = 0
        h_set = self.H_set[:, :, self.cur_step]
        self.fading = np.square(h_set) * self.path_loss
        return np.random.random((self.n_t * self.m_r, self.n_states))

    def sample(self):
        sample_action = np.random.randint(0, 10, self.n_t * self.m_r)
        return sample_action

    def cal_rate(self, power, fading):
        """Calculate channel rates.

        The receive power equals to emitting power times channel gain.
        The SINR can be calculated from all receive power. The channel rate
        can be infered by Shannon's formula.

        Args:
            power: emitting power.
            fading: channel gain infomation of current time slot.

        Returns:
            a vector of channel rate.
        """
        noise_power = 1e-3*pow(10., self.thres_power/10.)
        maxC = 1000.
        recv_power = power * fading

        signal_power = recv_power.diagonal()
        total_power = recv_power.sum(axis=1)
        inter_power = total_power - signal_power
        sinr = signal_power / (inter_power + noise_power)
        sinr = np.clip(sinr, 0, maxC)
        rate = np.log(1. + sinr)/np.log(2)

        return rate

    def get_state(self, power, rate, fading):
        """Calculate and constitute the state.

        The state on each receiving end consists of the following components:
        1. channel gain of the Rx, includes siginal and interferences
        2. Tx emitting power of all channel
        3. rate of all channel

        The m_state channels (except itself) are selected by the sorter
        assigned when initializing the observation space, default the
        emitting power.
        So there are m_state interferance channel gain, m_state Tx emitting
        power and m_state channel rate. Notice that Tx emitting power and
        channel rate should also include the infomation of this device.
        Which parts of the metrics will consist the state is assigned when
        initializing the observation space.

        Args:
            power: vector of emitting power of the last time slot.
            rate: vector of channel rate of the last time slot.
            fading: matrix of all channel gain of the last time slot.

        Returns:
            state consisted of assigned metrics ordered by assigned sorter.
        """
        n_t, m_r, n_bs, m_usr = self.n_t, self.m_r, self.n_bs, self.m_usr
        n_recvs = n_t * m_r + n_bs * m_usr
        m_state = self.m_state

        if m_state > n_recvs:
            msg = f"m_state should be less than n_recvs({n_recvs})" \
                f", but was {m_state}"
            raise ValueError(msg)

        # 将信号项提前
        rate_last = rate * np.ones([n_recvs, n_recvs])
        for i, _ in enumerate(rate_last):
            rate_last[i, 0], rate_last[i, i] = rate_last[i, i], rate_last[i, 0]
        power_last = power * np.ones([n_recvs, n_recvs])
        for i, _ in enumerate(power_last):
            power_last[i, 0], power_last[i, i] = power_last[i, i], \
                power_last[i, 0]
        ordered_fading = fading.copy()
        for i, _ in enumerate(ordered_fading):
            ordered_fading[i, 0], ordered_fading[i, i] = \
                ordered_fading[i, i], ordered_fading[i, 0]

        sinr_norm_fading = ordered_fading[:, 1:] / \
            np.tile(ordered_fading[:, 0:1], [1, n_recvs-1])
        sinr_norm_fading = np.log2(1. + sinr_norm_fading)

        sort_param = {
            'power': power_last[:, 1:],
            'rate': rate_last[:, 1:],
            'fading': sinr_norm_fading
        }

        # sorter = sinr_norm_inv
        sorter = sort_param[self.sorter]
        indices1 = np.tile(np.expand_dims(np.linspace(
            0, n_recvs-1, num=n_recvs, dtype=np.int32), axis=1), [1, m_state])
        indices2 = np.argsort(sorter, axis=1)[:, -m_state:]

        rate_last = np.hstack(
            [rate_last[:, 0:1], rate_last[indices1, indices2+1]])
        power_last = np.hstack(
            [power_last[:, 0:1], power_last[indices1, indices2+1]])
        sinr_norm_fading = sinr_norm_fading[indices1, indices2]
        metric_param = {
            'power': power_last,
            'rate': rate_last,
            'fading': sinr_norm_fading
        }
        state = np.hstack([metric_param[metric] for metric in self.metrics])

        return state[:self.n_t * self.m_r]

    def step(self, action, raw=False):
        h_set = self.H_set[:, :, self.cur_step]
        self.fading = np.square(h_set) * self.path_loss
        if raw:
            power = action
        else:
            power = self.power_levels[action]

        # append power of BS(constant)
        if len(power) == self.n_t * self.m_r:
            power = np.concatenate((power, np.full(self.m_usr, self.bs_power)))
        elif len(power) == self.n_recvs:
            power[self.n_t * self.m_r:] = np.full(self.m_usr, self.bs_power)
        else:
            msg = f"length of power should be n_recvs({self.n_recvs})" \
                f" or n_t*m_r({self.n_t*self.m_r}), but is {len(power)}"
            raise ValueError(msg)
        rate = self.cal_rate(power, self.fading)

        state = self.get_state(power, rate, self.fading)
        reward = np.sum(rate)
        done = self.cur_step == self.Ns - 1
        info = self.cur_step

        self.cur_step += 1

        return state, reward, done, info

    def render(self):

        def cir_edge(center, radius, color):
            patch = mpatches.Circle(center, radius,
                                    fc='white', ec=color, ls='--')
            return patch

        fig, ax = plt.subplots(figsize=(10, 10))
        ax.set_xlim(-1.2, 1.2)
        ax.set_ylim(-1.2, 1.2)

        # draw d2d pairs
        for t_idx, pair in self.devices.items():
            t, rs = pair['t_device'], pair['r_devices']
            # draw edge
            ax.add_patch(cir_edge((t.x, t.y), self.R_dev, 'green'))
            # draw t device
            ax.scatter([t.x], [t.y], marker='s', s=100, c='green', zorder=10)
            # draw r devices
            for _, r in rs.items():
                ax.scatter([r.x], [r.y], marker='o',
                           s=60, c='green', zorder=10)

        # draw cell and bs
        cell_xs = self.R_bs * \
            np.array([0, np.sqrt(3)/2, np.sqrt(3)/2,
                      0, -np.sqrt(3)/2, -np.sqrt(3)/2, 0])
        cell_ys = self.R_bs * np.array([1, .5, -.5, -1, -.5, .5, 1])
        ax.plot(cell_xs, cell_ys, color='black')

        ax.scatter([0.0], [0.0], marker='^', s=100, c='blue', zorder=30)
        # draw usrs
        for _, usr in self.users.items():
            ax.scatter([usr.x], [usr.y], marker='x',
                       s=100, c='orange', zorder=20)
            ax.plot([0, usr.x], [0, usr.y], ls='--', c='blue', zorder=20)

        save_path = Path('pa_env.png')
        plt.savefig(save_path)
        plt.close(fig)
        return save_path



from pa_env import (
    PAEnv,
    Node
)
import numpy as np
from pathlib import Path

users = {
    0: Node(0.1, 0, 'user'),
    1: Node(-0.1, 0, 'user'),
}
devices = {
    0: {
        't_device': Node(0, 0.5, 't_device'),
        'r_devices': {
            0: Node(0, 0.6, 'r_device')
        }
    },
    1: {
        't_device': Node(0, -0.5, 't_device'),
        'r_devices': {
            0: Node(0, -0.6, 'r_device')
        }
    }
}


def equal(unit, target):
    tolerance = 1e-6 * np.ones_like(target)
    return (np.abs(unit - target) < tolerance).all()


def test_init_power_level():
    """test transmit powers"""
    env = PAEnv(n_levels=4, min_power=10, max_power=30)
    power_levels = env.power_levels
    assert all(power_levels == [0., 0.01, 0.1, 1.])


def test_init_pos():
    """test position constraint"""
    env = PAEnv(n_levels=4)

    def dis(node, target):
        return np.sqrt(
            (node.x - target.x) ** 2 +
            (node.y - target.y) ** 2
        )
    # test bs users
    assert all(
        env.r_bs <= dis(usr, env.station) <= env.R_bs
        for usr in env.users.values()
    )

    # test devices
    for cluster in env.devices.values():
        t_device, r_devices = cluster['t_device'], cluster['r_devices']

        assert env.r_bs <= dis(t_device, env.station) <= (
            env.R_bs - env.R_dev)
        assert all(
            env.r_dev <= dis(r_device, t_device) <= env.R_dev
            for r_device in r_devices.values()
        )


def test_jakes():
    # TODO test stastic features of jakes
    # target_std, target_mean = 0.429, 1.253  # Rayleigh Distribution

    # x_len, y_len, Ns = H_set.shape
    # h_std = np.mean([
    #     H_set[x, y, :].std()
    #     for x in range(x_len)
    #     for y in range(y_len)
    # ])
    # assert (h_std - target_std) / target_std < 0.1

    # h_mean = np.mean([
    #     H_set[x, y, :].mean()
    #     for x in range(x_len)
    #     for y in range(y_len)
    # ])
    # assert (h_mean - target_mean) / target_mean < 0.05
    pass


def test_init_path_loss():
    """test distance, since lognormal is random"""
    env = PAEnv(n_levels=4, n_t_devices=2, m_r_devices=1, m_usrs=2)
    env.users = users
    env.devices = devices
    env.init_path_loss()
    distance_matrix = env.distance_matrix
    target_dis = np.array(
        [
            [0.1, 1.1, 0.6, 0.6],
            [1.1, 0.1, 0.6, 0.6],
            [np.sqrt(0.26), np.sqrt(0.26), 0.1, 0.1],
            [np.sqrt(0.26), np.sqrt(0.26), 0.1, 0.1],
        ]
    )
    assert equal(distance_matrix, target_dis)


def test_cal_rate():
    """test rate calc"""
    env = PAEnv(n_levels=4, n_t_devices=2, m_r_devices=1, m_usrs=2)
    power = [0.01, 0.01, 0.1, 0.1]
    fading = np.array([
        [1e-1, 1e-3, 1e-2, 1e-2],
        [1e-3, 1e-1, 1e-2, 1e-2],
        [1e-2, 1e-2, 1e-2, 1e-2],
        [1e-2, 1e-2, 1e-2, 1e-2],
    ])
    target_rate = np.array([0.58256799, 0.58256799, 0.87446912, 0.87446912])
    assert equal(env.cal_rate(power, fading), target_rate)


def test_get_state():
    env = PAEnv(n_levels=4, n_t_devices=4, m_r_devices=1, m_usrs=0)
    power = [0.01, 0.02, 0.03, 0.04]
    fading = np.array([
        [1.1e-1, 1.2e-3, 1.3e-2, 1.4e-2],
        [2.1e-3, 2.2e-1, 2.3e-2, 2.4e-2],
        [3.1e-2, 3.2e-2, 3.3e-2, 3.4e-2],
        [4.1e-2, 4.2e-2, 4.3e-2, 4.4e-2],
    ])
    rate = env.cal_rate(power, fading)
    # test error
    try:
        env.m_state = 8
        env.get_state(power, rate, fading)
    except Exception as e:
        assert e.__class__ == ValueError
        assert e.args[0] == 'm_state should be less than n_recvs(4)'\
            ', but was 8'

    # test value
    env.m_state = 2
    target_state = np.array(
        [[0.01, 0.03, 0.04,
          1.09042222, 0.51457317, 0.75950816,
          0.16115479, 0.1728366],
         [0.02, 0.03, 0.04,
          1.86122244, 0.51457317, 0.75950816,
          0.14345279, 0.14937762],
         [0.03, 0.02, 0.04,
          0.51457317, 1.86122244, 0.75950816,
          0.97797369, 1.02169507],
         [0.04, 0.02, 0.03,
          0.75950816, 1.86122244, 0.51457317,
          0.96683314, 0.98351188]]
    )
    assert equal(env.get_state(power, rate, fading), target_state)


def test_metrics():
    # test error
    try:
        env = PAEnv(n_levels=4, n_t_devices=4, m_r_devices=1,
                    m_usrs=0, metrics=["others"])
    except ValueError as e:
        assert e.args[0] == \
            "metrics should in power, rate and fading, but is ['others']"

    power = [0.01, 0.02, 0.03, 0.04]
    fading = np.array([
        [1.1e-1, 1.2e-3, 1.3e-2, 1.4e-2],
        [2.1e-3, 2.2e-1, 2.3e-2, 2.4e-2],
        [3.1e-2, 3.2e-2, 3.3e-2, 3.4e-2],
        [4.1e-2, 4.2e-2, 4.3e-2, 4.4e-2],
    ])
    # test power
    # one test is enough, case total combine metrics is tested in test_state
    env = PAEnv(n_levels=4, n_t_devices=4, m_r_devices=1, m_state=2,
                m_usrs=0, metrics=["power"])
    rate = env.cal_rate(power, fading)

    target_state = np.array(
        [[0.01, 0.03, 0.04, ],
         [0.02, 0.03, 0.04, ],
         [0.03, 0.02, 0.04, ],
         [0.04, 0.02, 0.03, ]]
    )
    assert equal(env.get_state(power, rate, fading), target_state)


def test_sorter():
    # test error
    try:
        env = PAEnv(n_levels=4, n_t_devices=4,
                    m_r_devices=1, m_usrs=0, sorter="others")
    except ValueError as e:
        assert e.args[0] == 'sorter should in power, rate'\
            ' and fading, but is others'

    # test power with power
    power = [0.04, 0.03, 0.02, 0.01]
    fading = np.array([
        [1.1e-1, 1.2e-3, 1.3e-2, 1.4e-2],
        [2.1e-3, 2.2e-1, 2.3e-2, 2.4e-2],
        [3.1e-2, 3.2e-2, 3.3e-2, 3.4e-2],
        [4.1e-2, 4.2e-2, 4.3e-2, 4.4e-2],
    ])
    env = PAEnv(n_levels=4, n_t_devices=4, m_r_devices=1, m_state=2,
                m_usrs=0, sorter="power", metrics=["power"])
    rate = env.cal_rate(power, fading)

    target_state = np.array(
        [[0.04, 0.02, 0.03, ],
         [0.03, 0.02, 0.04, ],
         [0.02, 0.03, 0.04, ],
         [0.01, 0.03, 0.04, ]]
    )
    assert equal(env.get_state(power, rate, fading), target_state)


def test_seed():
    env = PAEnv(n_levels=4, m_usrs=1, seed=123)
    # this is func in PAEnv to random pos

    def random_point(min_r, radius, ox=0, oy=0):
        theta = np.random.random() * 2 * np.pi
        r = np.random.uniform(min_r, radius**2)
        x, y = np.cos(theta) * np.sqrt(r), np.sin(theta) * np.sqrt(r)
        return ox + x, oy + y
    np.random.seed(123)
    target_x, target_y = random_point(env.r_bs, env.R_bs)
    usr = env.users[0]
    assert all((target_x == usr.x, target_y == usr.y))


def test_action():
    env = PAEnv(n_levels=10, seed=799345)
    n_actions = env.n_actions
    n_recvs = env.n_recvs
    # normal
    env.reset()
    np.random.seed(799345)
    action = np.random.randint(0, n_actions, (n_recvs, ))
    s_, r, d, i = env.step(action)
    assert r == 22.252017751938354
    # only D2D actions is enough
    env.reset()
    np.random.seed(799345)
    action = np.random.randint(0, n_actions, (n_recvs - env.m_usr, ))
    s_, r, d, i = env.step(action)
    assert r == 22.252017751938354
    # other action dim raises error
    env.reset()
    np.random.seed(799345)
    action = np.random.randint(0, n_actions, (n_recvs - env.m_usr - 1, ))
    try:
        s_, r, d, i = env.step(action)
    except ValueError as e:
        msg = f"length of power should be n_recvs({env.n_recvs})" \
            f" or n_t*m_r({env.n_t*env.m_r}), but is {len(action)}"
        assert e.args[0] == msg
    # raw
    env.reset()
    np.random.seed(799345)
    action = np.random.randint(0, n_actions, (n_recvs, ))
    raw_power = env.power_levels[action]
    s_, r, d, i = env.step(raw_power, raw=True)
    assert r == 22.252017751938354


def test_step():
    env = PAEnv(n_levels=10)
    n_actions, n_states = env.n_actions, env.n_states
    assert n_actions == 10
    assert n_states == 50
    env.reset()
    action = env.sample()
    env.step(action)
    fig: Path() = env.render()
    if fig.exists():
        fig.unlink()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 27 13:31:50 2019
@author: farismismar
"""
import gym
import math
from gym import spaces
from gym.utils import seeding
import numpy as np
from numpy import linalg as LA
import os
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
# The GPU id to use, usually either "0" or "1";
os.environ["CUDA_VISIBLE_DEVICES"] = "0"   # My NVIDIA GTX 1080 Ti FE GPU

# An attempt to follow
# https://github.com/openai/gym/blob/master/gym/envs/classic_control/cartpole.py

# Environment parameters
# cell radius
# UE movement speed
# BS max tx power
# BS antenna
# UE noise figure
# Center frequency
# Transmit antenna isotropic gain
# Antenna heights
# Shadow fading margin
# Number of ULA antenna elements
# Oversampling factor


class radio_environment(gym.Env):
    """In a two user two base station scenario, the SNR of the user depends on 
    the channel between the user and the base station, the transmission power and the beamforming mode of 
    the communication base station and the interference base station.
    
        Observation: 
            Num Observation                                    Min      Max
            0   User1 server X                                 -r       r
            1   User1 server Y                                 -r       r
            2   User2 server X                                 isd-r    isd+r
            3   User2 server Y                                 -r       r
            4   Serving BS Power                               5        40W
            5   Neighbor BS power                              5        40W
            6   BF codebook index for Serving                  0        M-1
            7   BF codebook index for Neighbor                 0        M-1

    Attributes:
        M_ULA: Base station employs a uniform linear array (ULA) of M antennas. 
        received_sinr_dB: SINR accepted by users.
        serving_transmit_power_dB: Transmission power of communication base station.
        interfering_transmit_power_dB: Transmission power of interfering base station. 
    """

    def __init__(self, seed):

        self.M_ULA = 4
        self.cell_radius = 150  # in meters.
        self.inter_site_distance = 3*self.cell_radius / 2.
        self.num_users = 30  # number of users.
        self.gamma_0 = 5  # beamforming constant SINR.
        self.min_sinr = -3  # in dB
        self.sinr_target = self.gamma_0 + 10*np.log10(self.M_ULA)  # in dB.
        self.max_tx_power = 40  # in Watts
        self.max_tx_power_interference = 40  # in Watts
        self.f_c = 28e9  # Hz
        self.G_ant_no_beamforming = 11  # dBi
        self.prob_LOS = 0.8  # Probability of LOS transmission
        self.num_actions = 16
        self.step_count = 0  # which step

        # Where are the base stations?
        self.x_bs_1, self.y_bs_1 = 0, 0
        self.x_bs_2, self.y_bs_2 = self.inter_site_distance, 0

        # for Beamforming
        self.use_beamforming = True
        self.k_oversample = 1  # oversampling factor
        self.Np = 4  # from 3 to 5 for mmWave
        self.F = np.zeros(
            [self.M_ULA, self.k_oversample * self.M_ULA], dtype=complex)
        self.theta_n = math.pi * \
            np.arange(start=0., stop=1., step=1. /
                      (self.k_oversample*self.M_ULA))

        # Beamforming codebook F
        for n in np.arange(self.k_oversample*self.M_ULA):
            f_n = self._compute_bf_vector(self.theta_n[n])
            self.F[:, n] = f_n
        self.f_n_bs1 = None  # The index in the codebook for serving BS
        self.f_n_bs2 = None  # The index in the codebook for interfering BS

        # for Reinforcement Learning
        self.reward_min = -20
        self.reward_max = 100

        bounds_lower = np.array([
            -self.cell_radius,
            -self.cell_radius,
            self.inter_site_distance-self.cell_radius,
            -self.cell_radius,
            1,
            1,
            0,
            0])

        bounds_upper = np.array([
            self.cell_radius,
            self.cell_radius,
            self.inter_site_distance+self.cell_radius,
            self.cell_radius,
            self.max_tx_power,
            self.max_tx_power_interference,
            self.k_oversample * self.M_ULA - 1,
            self.k_oversample * self.M_ULA - 1])

        self.action_space = spaces.Discrete(
            self.num_actions)  # action size is here
        # spaces.Discrete(2) # state size is here
        self.observation_space = spaces.Box(
            bounds_lower, bounds_upper, dtype=np.float32)

        self.seed(seed=seed)

        self.state = None
#        self.steps_beyond_done = None
        self.received_sinr_dB = None
        self.serving_transmit_power_dB = None
        self.interfering_transmit_power_dB = None

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(self, seed):
        """Reset enviornment.
        
        Args:
            seed: A fixed value.

        Returns:
            A fixed state.
            
        """
        # Initialize f_n of both cells
        self.np_random, seed = seeding.np_random(seed)
        self.f_n_bs1 = self.np_random.randint(self.M_ULA)
        self.f_n_bs2 = self.np_random.randint(self.M_ULA)

        self.state = [self.np_random.uniform(
                          low=-self.cell_radius, high=self.cell_radius),
                      self.np_random.uniform(
                          low=-self.cell_radius, high=self.cell_radius),
                      self.np_random.uniform(
                          low=self.inter_site_distance-self.cell_radius, high=self.inter_site_distance+self.cell_radius),
                      self.np_random.uniform(
                          low=-self.cell_radius, high=self.cell_radius),
                      self.np_random.uniform(
                          low=1, high=self.max_tx_power/2),
                      self.np_random.uniform(
                          low=1, high=self.max_tx_power_interference/2),
                      self.f_n_bs1,
                      self.f_n_bs2
                      ]

        self.step_count = 0

        return np.array(self.state)

    def step(self, action):
        """
        
        This function is used to update the environment after the agent outputs the action.
        
        Args:
            action: The action of agent.

        Returns:
            state: The state of doing an action.
            reward: Evaluate the result of an action.
            done: If the SNR, transmission power and interference power are within a certain 
                range after updating according to a certain action, done is true.
            abort: If the SNR, transmission power and interference power are not within a certain 
                range after updating according to a certain action, abort is true.
        """
       # assert self.action_space.contains(action), "%r (%s) invalid"%(action, type(action))

        state = self.state
        reward = 0
        x_ue_1, y_ue_1, x_ue_2, y_ue_2, pt_serving, pt_interferer, f_n_bs1, f_n_bs2 = state
        # based on the action make your call
        # only once a period, perform BF
        # The action is derived from a decimal interpretation
        ################################################################
        #    log_2 M (serving)    #    log_2 M (interferer)    # S # O #
        ################################################################

        if action != -1:  # optimal

            # int('0b0100101',2)
            power_command_l = action & 0b0001  # 1 power up, 0 power down
            # 1 power up, 0 power down
            power_command_b = (action & 0b0010) >> 1
            bf_selection_l = (action & 0b0100) >> 2  # 1 step up, 0 step down
            bf_selection_b = (action & 0b1000) >> 3

            self.step_count += 1
            if power_command_l == 0:
                pt_serving *= 10**(-1/10.)
            else:
                pt_serving *= 10**(1/10.)

            if power_command_b == 0:
                pt_interferer *= 10**(-1/10.)
            else:
                pt_interferer *= 10**(1/10.)

            if bf_selection_l == 1:
                f_n_bs1 = (f_n_bs1 + 1) % self.k_oversample * self.M_ULA
            else:
                f_n_bs1 = (f_n_bs1 - 1) % self.k_oversample * self.M_ULA

            if bf_selection_b == 1:
                f_n_bs2 = (f_n_bs2 + 1) % self.k_oversample * self.M_ULA
            else:
                f_n_bs2 = (f_n_bs2 - 1) % self.k_oversample * self.M_ULA

        elif action > self.num_actions - 1:
            print('WARNING: Invalid action played!')
            reward = 0
            return [], 0, False, True

        # move the UEs at a speed of v, in a random direction
        v = 2  # km/h.

        v *= 5./18  # in m/sec
        theta_1, theta_2 = self.np_random.uniform(
            low=-math.pi, high=math.pi, size=2)

        dx_1 = v * math.cos(theta_1)
        dy_1 = v * math.sin(theta_1)

        dx_2 = v * math.cos(theta_2)
        dy_2 = v * math.sin(theta_2)

        # Move UE 1
        x_ue_1 += dx_1
        y_ue_1 += dy_1

        # Move UE 2
        x_ue_2 += dx_2
        y_ue_2 += dy_2

        # Update the beamforming codebook index
        self.f_n_bs1 = f_n_bs1
        self.f_n_bs2 = f_n_bs2

        received_power, interference_power, received_sinr = self._compute_rf(
            x_ue_1, y_ue_1, pt_serving, pt_interferer, is_ue_2=False)
        received_power_ue2, interference_power_ue2, received_ue2_sinr = self._compute_rf(
            x_ue_2, y_ue_2, pt_serving, pt_interferer, is_ue_2=True)

        # keep track of quantities...
        self.received_sinr_dB = received_sinr
        self.received_ue2_sinr_dB = received_ue2_sinr
        self.serving_transmit_power_dBm = 10*np.log10(pt_serving*1e3)
        self.interfering_transmit_power_dBm = 10*np.log10(pt_interferer*1e3)

        # Did we find a FEASIBLE NON-DEGENERATE solution?
        done = all([(pt_serving <= self.max_tx_power), (pt_serving >= 0), (pt_interferer <= self.max_tx_power_interference), 
            (pt_interferer >= 0), (received_sinr >= self.min_sinr), (received_ue2_sinr >= self.min_sinr), 
            (received_sinr >= self.sinr_target), (received_ue2_sinr >= self.sinr_target)])

        abort = any([(pt_serving > self.max_tx_power), (pt_interferer > self.max_tx_power_interference), (received_sinr < self.min_sinr), 
            (received_ue2_sinr < self.min_sinr), (received_sinr > 70), (received_ue2_sinr > 70)])

#        print('{:.2f} dB | {:.2f} dB | {:.2f} W | {:.2f} W '.format(received_sinr, received_ue2_sinr, pt_serving, pt_interferer), end='')
#        print('Done: {}'.format(done))
#        print('UE moved to ({0:0.3f},{1:0.3f}) and their received SINR became {2:0.3f} dB.'.format(x,y,received_sinr))

        # the reward
        reward = received_sinr + received_ue2_sinr

        # Update the state.
        self.state = (x_ue_1, y_ue_1, x_ue_2, y_ue_2, pt_serving,
                      pt_interferer, f_n_bs1, f_n_bs2)

        if abort == True:
            done = False
            reward = self.reward_min
        else:
            reward += self.reward_max

#        print(done, (received_sinr >= self.sinr_target) , (pt_serving <= self.max_tx_power) , (pt_serving >= 0) , \
#                (pt_interferer <= self.max_tx_power_interference) , (pt_interferer >= 0) , (received_ue2_sinr >= self.sinr_target))

        if action == -1:        # for optimal
            return np.array(self.state), reward, False, False

        return np.array(self.state), reward, done, abort

    def _compute_bf_vector(self, theta):
        """This function corresponds to the formula in the paper (2).

        Args:
            theta: steering angle.
            
        Returns:
            f: Beamforming Vectors

        """
        c = 3e8  # speed of light
        wavelength = c / self.f_c

        d = wavelength / 2.  # antenna spacing
        k = 2. * math.pi / wavelength

        exponent = 1j * k * d * math.cos(theta) * np.arange(self.M_ULA)

        f = 1. / math.sqrt(self.M_ULA) * np.exp(exponent)

        # Test the norm square... is it equal to unity? YES.
    #    norm_f_sq = LA.norm(f, ord=2) ** 2
     #   print(norm_f_sq)

        return f

    def _compute_channel(self, x_ue, y_ue, x_bs, y_bs):
        """Random value is included in the formula, and formula test is not conducted.
        
        Corresponding to the formula (3) in the paper.

        Args:
        x_ue: The abscissa values of user.
        y_ue: The ordinate values of user.
        x_bs: The abscissa values of base station. 
        y_bs: The ordinate values of base station. 

        Returns: Channel model of downlink.

        """
        # Np is the number of paths p
        PLE_L = 2
        PLE_N = 4
        G_ant = 3  # dBi for beamforming mmWave antennas

        # Override the antenna gain if no beamforming
        if self.use_beamforming == False:
            G_ant = self.G_ant_no_beamforming

        # theta is the steering angle.  Sampled iid from unif(0,pi).
        theta = np.random.uniform(low=0, high=math.pi, size=self.Np)

        is_mmWave = (self.f_c > 25e9)
        # Calculate the transmission path loss of LOS and NLOS.
        if is_mmWave:
            path_loss_LOS = 10 ** (self._path_loss_mmWave(x_ue,
                                                          y_ue, PLE_L, x_bs, y_bs) / 10.)
            path_loss_NLOS = 10 ** (self._path_loss_mmWave(x_ue,
                                                           y_ue, PLE_N, x_bs, y_bs) / 10.)
        else:
            path_loss_LOS = 10 ** (self._path_loss_sub6(x_ue,
                                                        y_ue, x_bs, y_bs) / 10.)
            path_loss_NLOS = 10 ** (self._path_loss_sub6(x_ue,
                                                         y_ue, x_bs, y_bs) / 10.)

        # Bernoulli for p
        alpha = np.zeros(self.Np, dtype=complex)
        p = np.random.binomial(1, self.prob_LOS)
        # 80% probability is transmitted by Los, and 20% probability is transmitted by NLOS.
        
        # For Los transmission, the number of transmission channels is only 1.
        if p == 1:
            self.Np = 1
            alpha[0] = 1. / math.sqrt(path_loss_LOS)
        else:
            # just changed alpha to be complex in the case of NLOS
            # The NLOS gain follows Rayleigh distribution and 
            # the number of transmission channels is 4.
            alpha = (np.random.normal(size=self.Np) + 1j *
                     np.random.normal(size=self.Np)) / math.sqrt(path_loss_NLOS)

        rho = 1. * 10 ** (G_ant / 10.)
        # The reduction of gain loss caused by beamforming.
        # initialize the channel as a complex variable.
        h = np.zeros(self.M_ULA, dtype=complex)

        for p in np.arange(self.Np):
            a_theta = self._compute_bf_vector(theta[p])
            # scalar multiplication into a vector
            h += alpha[p] / rho * a_theta.T

        h *= math.sqrt(self.M_ULA)

#        print ('Warning: channel gain is {} dB.'.format(10*np.log10(LA.norm(h, ord=2))))
        return h

    def _compute_rf(self, x_ue, y_ue, pt_bs1, pt_bs2, is_ue_2=False):
        """Random value is included in the formula, and formula test is not conducted.
        
        Corresponding to the formula (4) and (5) in the paper.
        
        Args：
        x_ue: The abscissa values of user.
        y_ue: The ordinate values of user.
        pt_bs1: Transmit power of base station 1.
        pt_bs2: Transmit power of base station 2.
        is_ue_2: 
            The default is False.
            If it is True, then base station 2 is communicating with the user, 
            and the signal sent by base station 1 is interference signal.
            If it is False, then base station 1 is communicating with the user, 
            and the signal sent by base station 2 is interference signal.

        Returns:
        received_power: Useful power received by user.
        interference_power: Interference power received by user.
        received_sinr: User's SNR (downlink SNR).

        """

        T = 290  # Kelvins
        B = 15000  # Hz
        k_Boltzmann = 1.38e-23

        noise_power = k_Boltzmann*T*B  # this is in Watts

        if is_ue_2 == False:
            # Without loss of generality, the base station is at the origin
            # The interfering base station is x = cell_radius, y = 0
            x_bs_1, y_bs_1 = self.x_bs_1, self.y_bs_1
            x_bs_2, y_bs_2 = self.x_bs_2, self.y_bs_2

            # Now the channel h, which is a vector in beamforming.
            # This computes the channel for user in serving BS from the serving BS.
            h_1 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_1, y_bs=y_bs_1)
            # This computes the channel for user in serving BS from the interfering BS.
            h_2 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_2, y_bs=y_bs_2)

            # if this is not beamforming, there is no precoder:
            # Corresponding to the formula (4) in the paper.
            if self.use_beamforming:
                received_power = pt_bs1 * \
                    abs(np.dot(h_1.conj(), self.F[:, self.f_n_bs1])) ** 2
                interference_power = pt_bs2 * \
                    abs(np.dot(h_2.conj(), self.F[:, self.f_n_bs2])) ** 2
            else:  # the gain is ||h||^2
                received_power = pt_bs1 * LA.norm(h_1, ord=2) ** 2
                interference_power = pt_bs2 * LA.norm(h_2, ord=2) ** 2
        else:
            x_bs_1, y_bs_1 = self.x_bs_1, self.y_bs_1
            x_bs_2, y_bs_2 = self.x_bs_2, self.y_bs_2

            # Now the channel h, which is a vector in beamforming.
            # This computes the channel for user in serving BS from the serving BS.
            h_1 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_2, y_bs=y_bs_2)
            # This computes the channel for user in serving BS from the interfering BS.
            h_2 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_1, y_bs=y_bs_1)

            # if this is not beamforming, there is no precoder:
            if self.use_beamforming:
                received_power = pt_bs2 * \
                    abs(np.dot(h_1.conj(), self.F[:, self.f_n_bs2])) ** 2
                interference_power = pt_bs1 * \
                    abs(np.dot(h_2.conj(), self.F[:, self.f_n_bs1])) ** 2
            else:  # the gain is ||h||^2
                received_power = pt_bs2 * LA.norm(h_1, ord=2) ** 2
                interference_power = pt_bs1 * LA.norm(h_2, ord=2) ** 2

        # Corresponding to the formula (5) in the paper.
        interference_plus_noise_power = interference_power + noise_power
        received_sinr = 10 * \
            np.log10(received_power / interference_plus_noise_power)

        return [received_power, interference_power, received_sinr]

    # https://ieeexplore-ieee-org.ezproxy.lib.utexas.edu/stamp/stamp.jsp?tp=&arnumber=7522613
    def _path_loss_mmWave(self, x, y, PLE, x_bs=0, y_bs=0):
        """Normal distribution is included in the formula, and formula test is not conducted.

        Args:
        x : The abscissa values of user.
        y : The ordinate values of user.
        PLE: path loss exponent.
        x_bs: The abscissa values of base station. The default is 0.
        y_bs: The abscissa values of base station. The default is 0.

        Returns:
        L: Transmission loss of millimeter wave channel.

        """

        # These are the parameters for f = 28000 MHz.
        c = 3e8  # speed of light
        wavelength = c / self.f_c
        A = 0.0671
        Nr = self.M_ULA
        sigma_sf = 9.1
        #PLE = 3.812

        d = math.sqrt((x - x_bs)**2 + (y - y_bs)**2)  # in meters
        if d == 0:
            L = 0
        else:
            fspl = 10 * np.log10(((4*math.pi*d) / wavelength) ** 2)
            pl = fspl + 10 * np.log10(d ** PLE) * (1 - A*np.log2(Nr))

            chi_sigma = np.random.normal(0, sigma_sf)  # log-normal shadowing
            L = pl + chi_sigma

        return L  # in dB

    def _path_loss_sub6(self, x, y, x_bs=0, y_bs=0):
        """
        
        Args:
        x: The abscissa values of user.
        y: The ordinate values of user
        x_bs: The abscissa values of base station. The default is 0.
        y_bs: The ordinate values of base station. The default is 0.

        Returns:
        L: Transmission loss of channel.

        """
        f_c = self.f_c
        c = 3e8  # speed of light
        d = math.sqrt((x - x_bs)**2 + (y - y_bs)**2)
        if d == 0:
            L = 0
        else:
            h_B = 20
            h_R = 1.5

    #        print('Distance from cell site is: {} km'.format(d/1000.))
            # FSPL
            L_fspl = -10*np.log10((4.*math.pi*c/f_c / d) ** 2)

            # COST-231Hata model. It is a empirical formula.
            C = 3#Correction factor of metropolis center
            a = (1.1 * np.log10(f_c/1e6) - 0.7) * \
                h_R - (1.56*np.log10(f_c/1e6) - 0.8)
            L_cost231 = 46.3 + 33.9 * np.log10(f_c/1e6) + 13.82 * np.log10(
                h_B) - a + (44.9 - 6.55 * np.log10(h_B)) * np.log10(d/1000.) + C

            L = L_cost231

        return L  # in dB








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 27 13:31:50 2019
@author: farismismar
"""

import math
from gym import spaces
from gym.utils import seeding
import numpy as np
from numpy import linalg as LA

# An attempt to follow
# https://github.com/openai/gym/blob/master/gym/envs/classic_control/cartpole.py

# Environment parameters
# cell radius
# UE movement speed
# BS max tx power
# BS antenna
# UE noise figure
# Center frequency
# Transmit antenna isotropic gain
# Antenna heights
# Shadow fading margin
# Number of ULA antenna elements
# Oversampling factor

class radio_environment:
    '''    
        Observation: 
            Type: Box(6 or 8)
            Num Observation                                    Min      Max
            0   User1 server X                                 -r       r
            1   User1 server Y                                 -r       r
            2   User2 server X                                 isd-r    isd+r
            3   User2 server Y                                 -r       r
            4   Serving BS Power                               5        40W
            5   Neighbor BS power                              5        40W
            6   BF codebook index for Serving                  0        M-1
            7   BF codebook index for Neighbor                 0        M-1
                       
    '''     
    def __init__(self, seed):
        self.M_ULA = 4
                
        self.cell_radius = 150 # in meters.
        self.inter_site_distance = 3 * self.cell_radius / 2.
        self.num_users = 30 # number of users.
        self.gamma_0 = 5 # beamforming constant SINR.

        self.min_sinr = -3 # in dB
        self.sinr_target = self.gamma_0 + 10*np.log10(self.M_ULA) # in dB.
        self.max_tx_power = 40 # in Watts
        self.max_tx_power_interference = 40 # in Watts
        self.f_c = 28e9 # Hz
        self.G_ant_no_beamforming = 11 # dBi
        self.prob_LOS = 0.8 # Probability of LOS transmission

        self.num_actions = 16

        self.step_count = 0 # which step


        # Where are the base stations?
        self.x_bs_1, self.y_bs_1 = 0, 0
        self.x_bs_2, self.y_bs_2 = self.inter_site_distance, 0
        
        # for Beamforming
        self.use_beamforming = True
        self.k_oversample = 1 # oversampling factor
        self.Np = 4 # from 3 to 5 for mmWave
        self.F = np.zeros([self.M_ULA, self.k_oversample*self.M_ULA], dtype=complex)
        self.theta_n = math.pi * np.arange(start=0., stop=1., step=1./(self.k_oversample*self.M_ULA))
        # Beamforming codebook F
        for n in np.arange(self.k_oversample*self.M_ULA):
            f_n = self._compute_bf_vector(self.theta_n[n])
            self.F[:,n] = f_n
        self.f_n_bs1 = None  # The index in the codebook for serving BS
        self.f_n_bs2 = None  # The index in the codebook for interfering BS

        # for Reinforcement Learning
        self.reward_min = -20
        self.reward_max = 100
        
        bounds_lower = np.array([
            -self.cell_radius,
            -self.cell_radius,
            self.inter_site_distance-self.cell_radius,
            -self.cell_radius,
            1,
            1,
            0,
            0])

        bounds_upper = np.array([
            self.cell_radius,
            self.cell_radius,
            self.inter_site_distance+self.cell_radius,
            self.cell_radius,
            self.max_tx_power,
            self.max_tx_power_interference,
            self.k_oversample*self.M_ULA - 1,
            self.k_oversample*self.M_ULA - 1])

        self.action_space = spaces.Discrete(self.num_actions) # action size is here
        self.observation_space = spaces.Box(bounds_lower, bounds_upper, dtype=np.float32) # spaces.Discrete(2) # state size is here 
        
        self.seed(seed=seed)
        
        self.state = None
#        self.steps_beyond_done = None
        self.received_sinr_dB = None
        self.serving_transmit_power_dB = None
        self.interfering_transmit_power_dB = None
      
    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]
        
    def reset(self,seed):
        # Initialize f_n of both cells
        self.np_random, seed = seeding.np_random(seed)
        self.f_n_bs1 = self.np_random.randint(self.M_ULA)
        self.f_n_bs2 = self.np_random.randint(self.M_ULA)
        
        self.state = [self.np_random.uniform(low=-self.cell_radius, high=self.cell_radius),
                      self.np_random.uniform(low=-self.cell_radius, high=self.cell_radius),
                      self.np_random.uniform(low=self.inter_site_distance-self.cell_radius, high=self.inter_site_distance+self.cell_radius),
                      self.np_random.uniform(low=-self.cell_radius, high=self.cell_radius),
                      self.np_random.uniform(low=1, high=self.max_tx_power/2),
                      self.np_random.uniform(low=1, high=self.max_tx_power_interference/2),
                      self.f_n_bs1,
                      self.f_n_bs2
                      ]

        self.step_count = 0

        return np.array(self.state)
    
    def step(self, action):
       # assert self.action_space.contains(action), "%r (%s) invalid"%(action, type(action))

        state = self.state
        reward = 0
        x_ue_1, y_ue_1, x_ue_2, y_ue_2, pt_serving, pt_interferer, f_n_bs1, f_n_bs2 = state
        # based on the action make your call
        # only once a period, perform BF
        # The action is derived from a decimal interpretation
        ################################################################
        #    log_2 M (serving)    #    log_2 M (interferer)    # S # O #
        ################################################################
        
        if (action != -1): # optimal
    
            # int('0b0100101',2)
            power_command_l = action & 0b0001 # 1 power up, 0 power down
            power_command_b = (action & 0b0010) >> 1 # 1 power up, 0 power down
            bf_selection_l = (action & 0b0100) >> 2 # 1 step up, 0 step down
            bf_selection_b = (action & 0b1000) >> 3
            
            self.step_count += 1                
            if (power_command_l == 0):
                pt_serving *= 10**(-0.03/10.)
            else:
                pt_serving *= 10**(0.03/10.)
                
            if (power_command_b == 0):
                pt_interferer *= 10**(-0.03/10.)
            else:
                pt_interferer *= 10**(0.03/10.)
                
            if (bf_selection_l == 1):
                f_n_bs1 = (f_n_bs1 + 1) % self.k_oversample*self.M_ULA
            else:
                f_n_bs1 = (f_n_bs1 - 1) % self.k_oversample*self.M_ULA
    
            if (bf_selection_b == 1):
                f_n_bs2 = (f_n_bs2 + 1) % self.k_oversample*self.M_ULA
            else:
                f_n_bs2 = (f_n_bs2 - 1) % self.k_oversample*self.M_ULA
                                
        elif (action > self.num_actions - 1):
            print('WARNING: Invalid action played!')
            reward = 0
            return [], 0, False, True    
        
        # move the UEs at a speed of v, in a random direction
        v = 2 # km/h.

        v *= 5./18 # in m/sec
        theta_1, theta_2 = self.np_random.uniform(low=-math.pi, high=math.pi, size=2)
        
        dx_1 = v * math.cos(theta_1)
        dy_1 = v * math.sin(theta_1)

        dx_2 = v * math.cos(theta_2)
        dy_2 = v * math.sin(theta_2)
        
        # Move UE 1
        x_ue_1 += dx_1
        y_ue_1 += dy_1
        
        # Move UE 2
        x_ue_2 += dx_2
        y_ue_2 += dy_2
        
        # Update the beamforming codebook index
        self.f_n_bs1 = f_n_bs1
        self.f_n_bs2 = f_n_bs2
                
        received_power, interference_power, received_sinr = self._compute_rf(x_ue_1, y_ue_1, pt_serving, pt_interferer, is_ue_2=False)
        received_power_ue2, interference_power_ue2, received_ue2_sinr = self._compute_rf(x_ue_2, y_ue_2, pt_serving, pt_interferer, is_ue_2=True)
            
        # keep track of quantities...
        self.received_sinr_dB = received_sinr 
        self.received_ue2_sinr_dB = received_ue2_sinr
        self.serving_transmit_power_dBm = 10*np.log10(pt_serving*1e3)
        self.interfering_transmit_power_dBm = 10*np.log10(pt_interferer*1e3)

        # Did we find a FEASIBLE NON-DEGENERATE solution?
        done = (pt_serving <= self.max_tx_power) and (pt_serving >= 0) and (pt_interferer <= self.max_tx_power_interference) and (pt_interferer >= 0) and \
                (received_sinr >= self.min_sinr) and (received_ue2_sinr >= self.min_sinr) and (received_sinr >= self.sinr_target) and (received_ue2_sinr >= self.sinr_target)

              
        abort = (pt_serving > self.max_tx_power) or (pt_interferer > self.max_tx_power_interference) or (received_sinr < self.min_sinr) or (received_ue2_sinr < self.min_sinr) \
            or (received_sinr > 70) or (received_ue2_sinr > 70) #or (received_sinr < 10) or (received_ue2_sinr < 10)  
                
#        print('{:.2f} dB | {:.2f} dB | {:.2f} W | {:.2f} W '.format(received_sinr, received_ue2_sinr, pt_serving, pt_interferer), end='')
#        print('Done: {}'.format(done))
#        print('UE moved to ({0:0.3f},{1:0.3f}) and their received SINR became {2:0.3f} dB.'.format(x,y,received_sinr))
        
        # the reward
        reward = received_sinr + received_ue2_sinr

        # Update the state.        
        self.state = (x_ue_1, y_ue_1, x_ue_2, y_ue_2, pt_serving, pt_interferer, f_n_bs1, f_n_bs2)
     
        if abort == True:
            done = False
            reward = self.reward_min
        elif done:
            reward += self.reward_max

#        print(done, (received_sinr >= self.sinr_target) , (pt_serving <= self.max_tx_power) , (pt_serving >= 0) , \
#                (pt_interferer <= self.max_tx_power_interference) , (pt_interferer >= 0) , (received_ue2_sinr >= self.sinr_target))
        
        if action == -1:        # for optimal
            return np.array(self.state), reward, False, False
                
        return np.array(self.state), reward, done, abort

    def _compute_bf_vector(self, theta):
        c = 3e8 # speed of light
        wavelength = c / self.f_c
        
        d = wavelength / 2. # antenna spacing 
        k = 2. * math.pi / wavelength
    
        exponent = 1j * k * d * math.cos(theta) * np.arange(self.M_ULA)
        
        f = 1. / math.sqrt(self.M_ULA) * np.exp(exponent)
        
        # Test the norm square... is it equal to unity? YES.
    #    norm_f_sq = LA.norm(f, ord=2) ** 2
     #   print(norm_f_sq)
    
        return f

    def _compute_channel(self, x_ue, y_ue, x_bs, y_bs):
        # Np is the number of paths p
        PLE_L = 2
        PLE_N = 4
        G_ant = 3 # dBi for beamforming mmWave antennas
        
        # Override the antenna gain if no beamforming
        if self.use_beamforming == False:
            G_ant = self.G_ant_no_beamforming
            
        # theta is the steering angle.  Sampled iid from unif(0,pi).
        theta = np.random.uniform(low=0, high=math.pi, size=self.Np)
    
        is_mmWave = (self.f_c > 25e9)
        
        if is_mmWave:
            path_loss_LOS = 10 ** (self._path_loss_mmWave(x_ue, y_ue, PLE_L, x_bs, y_bs) / 10.)
            path_loss_NLOS = 10 ** (self._path_loss_mmWave(x_ue, y_ue, PLE_N, x_bs, y_bs) / 10.)
        else:
            path_loss_LOS = 10 ** (self._path_loss_sub6(x_ue, y_ue, x_bs, y_bs) / 10.)
            path_loss_NLOS = 10 ** (self._path_loss_sub6(x_ue, y_ue, x_bs, y_bs) / 10.)
            
        # Bernoulli for p
        alpha = np.zeros(self.Np, dtype=complex)
        p = np.random.binomial(1, self.prob_LOS)
        
        if (p == 1):
            self.Np = 1
            alpha[0] = 1. / math.sqrt(path_loss_LOS)
        else:
            ## just changed alpha to be complex in the case of NLOS
            alpha = (np.random.normal(size=self.Np) + 1j * np.random.normal(size=self.Np)) / math.sqrt(path_loss_NLOS)
                
        rho = 1. * 10 ** (G_ant / 10.)
        
        # initialize the channel as a complex variable.
        h = np.zeros(self.M_ULA, dtype=complex)
        
        for p in np.arange(self.Np):
            a_theta = self._compute_bf_vector(theta[p])
            h += alpha[p] / rho * a_theta.T # scalar multiplication into a vector
        
        h *= math.sqrt(self.M_ULA)
        
#        print ('Warning: channel gain is {} dB.'.format(10*np.log10(LA.norm(h, ord=2))))
        return h

    def _compute_rf(self, x_ue, y_ue, pt_bs1, pt_bs2, is_ue_2=False):
        T = 290 # Kelvins
        B = 15000 # Hz
        k_Boltzmann = 1.38e-23
        
        noise_power = k_Boltzmann*T*B # this is in Watts

        if is_ue_2 == False:
            # Without loss of generality, the base station is at the origin
            # The interfering base station is x = cell_radius, y = 0
            x_bs_1, y_bs_1 = self.x_bs_1, self.y_bs_1
            x_bs_2, y_bs_2 = self.x_bs_2, self.y_bs_2

            # Now the channel h, which is a vector in beamforming.
            # This computes the channel for user in serving BS from the serving BS.
            h_1 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_1, y_bs=y_bs_1) 
            # This computes the channel for user in serving BS from the interfering BS.
            h_2 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_2, y_bs=y_bs_2)
              
            # if this is not beamforming, there is no precoder:
            if (self.use_beamforming):
                received_power = pt_bs1 * abs(np.dot(h_1.conj(), self.F[:, self.f_n_bs1])) ** 2
                interference_power = pt_bs2 * abs(np.dot(h_2.conj(), self.F[:, self.f_n_bs2])) ** 2
            else: # the gain is ||h||^2
                received_power = pt_bs1 * LA.norm(h_1, ord=2) ** 2
                interference_power = pt_bs2 * LA.norm(h_2, ord=2) ** 2
        else:
            x_bs_1, y_bs_1 = self.x_bs_1, self.y_bs_1
            x_bs_2, y_bs_2 = self.x_bs_2, self.y_bs_2            
            
            # Now the channel h, which is a vector in beamforming.
            # This computes the channel for user in serving BS from the serving BS.
            h_1 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_2, y_bs=y_bs_2) 
            # This computes the channel for user in serving BS from the interfering BS.
            h_2 = self._compute_channel(x_ue, y_ue, x_bs=x_bs_1, y_bs=y_bs_1) 

            # if this is not beamforming, there is no precoder:
            if (self.use_beamforming):
                received_power = pt_bs2 * abs(np.dot(h_1.conj(), self.F[:, self.f_n_bs2])) ** 2
                interference_power = pt_bs1 * abs(np.dot(h_2.conj(), self.F[:, self.f_n_bs1])) ** 2
            else: # the gain is ||h||^2
                received_power = pt_bs2 * LA.norm(h_1, ord=2) ** 2
                interference_power = pt_bs1 * LA.norm(h_2, ord=2) ** 2
                
        interference_plus_noise_power = interference_power + noise_power
        received_sinr = 10*np.log10(received_power / interference_plus_noise_power)

        return [received_power, interference_power, received_sinr]
    
    # https://ieeexplore-ieee-org.ezproxy.lib.utexas.edu/stamp/stamp.jsp?tp=&arnumber=7522613
    def _path_loss_mmWave(self, x, y, PLE, x_bs=0, y_bs=0):
        # These are the parameters for f = 28000 MHz.
        c = 3e8 # speed of light
        wavelength = c / self.f_c
        A = 0.0671
        Nr = self.M_ULA
        sigma_sf = 9.1
        #PLE = 3.812
        
        d = math.sqrt((x - x_bs)**2 + (y - y_bs)**2) # in meters
        
        fspl = 10 * np.log10(((4*math.pi*d) / wavelength) ** 2)
        pl = fspl + 10 * np.log10(d ** PLE) * (1 - A*np.log2(Nr))
    
        chi_sigma = np.random.normal(0, sigma_sf) # log-normal shadowing 
        L = pl + chi_sigma
    
        return L # in dB    
        
    def _path_loss_sub6(self, x, y, x_bs=0, y_bs=0):
        f_c = self.f_c
        c = 3e8 # speed of light
        d = math.sqrt((x - x_bs)**2 + (y - y_bs)**2)
        h_B = 20
        h_R = 1.5

#        print('Distance from cell site is: {} km'.format(d/1000.))
        # FSPL
        L_fspl = -10*np.log10((4.*math.pi*c/f_c / d) ** 2)
        
        # COST231        
        C = 3
        a = (1.1 * np.log10(f_c/1e6) - 0.7)*h_R - (1.56*np.log10(f_c/1e6) - 0.8)
        L_cost231  = 46.3 + 33.9 * np.log10(f_c/1e6) + 13.82 * np.log10(h_B) - a + (44.9 - 6.55 * np.log10(h_B)) * np.log10(d/1000.) + C
    
        L = L_cost231
        
        return L # in dB



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 16 15:15:30 2020

@author: yhh18
@desc: 单元测试environment
"""

from PC_env import radio_environment
import pytest
import numpy as np


class TestEnvFunc:

    def __init__(self):
        self.seed = 2  # Seed can take any value between 0 and 49
        with pytest.warns(None) as warnings:
            self.env = radio_environment(self.seed)

    # Check that dtype is explicitly declared for gym.Box spaces
        for warning_msg in warnings:
            assert not 'autodetected dtype' in str(warning_msg.message)

        ob_space = self.env.observation_space
        act_space = self.env.action_space
        ob = self.env.reset(self.seed)
        assert ob_space.contains(
            ob), 'Reset observation: {!r} not in space'.format(ob)
        action = act_space.sample()
        observation, reward, done, _info = self.env.step(action)
        assert ob_space.contains(
            observation), 'Step observation: {!r} not in space'.format(observation)
        assert np.isscalar(reward), "{} is not a scalar for {}".format(
            reward, self.env)
        assert isinstance(
            done, bool), "Expected {} to be a boolean".format(done)

    def test_reset(self):
        # Reset to the same state every time.
        state_1 = self.env.reset(self.seed)
        state_2 = self.env.reset(self.seed)
        assert not (state_2 - state_1).any()

    def test_step(self):
        # This function calls almost any other function in the ns_environment.

        for action in range(self.env.num_actions):
            self.env.reset(self.seed)
            next_observation, reward, done, abort = self.env.step(action)
            assert next_observation is not None
            assert reward is not None
            assert done is not None
            assert abort is not None
